@@ -2,7 +2,7 @@
 
 package Git::Hooks::CheckFile;
 # ABSTRACT: Git::Hooks plugin for checking files
-$Git::Hooks::CheckFile::VERSION = '2.9.2';
+$Git::Hooks::CheckFile::VERSION = '2.9.3';
 use 5.010;
 use utf8;
 use strict;
@@ -33,7 +33,7 @@ sub _setup_config {
 }
 
 sub check_command {
-    my ($git, $commit, $file, $command) = @_;
+    my ($git, $ctx, $commit, $file, $command) = @_;
 
     my $tmpfile = $git->blob($commit, $file)
         or return;
@@ -81,7 +81,7 @@ sub check_command {
         # $file to avoid confounding the user.
         $output =~ s/\Q$tmpfile\E/$file/g;
 
-        $git->fault($message, {commit => $commit, details => $output});
+        $git->fault($message, {%$ctx, details => $output});
         return;
     } else {
         # FIXME: What should we do with eventual output from a
@@ -91,9 +91,9 @@ sub check_command {
 }
 
 sub check_new_files {
-    my ($git, $ctx, $commit, $name_status) = @_;
+    my ($git, $ctx, $commit, $name2status) = @_;
 
-    return 0 unless %$name_status; # No new file to check
+    return 0 unless %$name2status; # No new file to check
 
     # Construct a list of command checks from the
     # githooks.checkfile.name configuration. Each check in the list is a
@@ -143,9 +143,9 @@ sub check_new_files {
     my $errors = 0;
 
   FILE:
-    foreach my $file (sort keys %$name_status) {
+    foreach my $file (sort keys %$name2status) {
         # We're only interested in new and modified files
-        next unless $name_status->{$file} =~ /[AM]/;
+        next unless $name2status->{$file} =~ /[AM]/;
 
         my $basename = path($file)->basename;
 
@@ -191,7 +191,7 @@ EOS
         }
 
         foreach my $command (map {$_->[1]} grep {$basename =~ $_->[0]} @name_checks) {
-            check_command($git, $commit, $file, $command)
+            check_command($git, $ctx, $commit, $file, $command)
                 or ++$errors;
         }
 
@@ -232,18 +232,21 @@ EOS
 }
 
 sub deny_case_conflicts {
-    my ($git, $ctx, $commit, $name_status) = @_;
+    my ($git, $ctx, $commit, $name2status) = @_;
 
     return 0 unless $git->get_config_boolean($CFG => 'deny-case-conflict');
 
     # We're only interested in new file names
-    my @names = sort grep {$name_status->{$_} =~ /[AC]/} keys %$name_status;
+    my @names = sort grep {$name2status->{$_} =~ /[AC]/} keys %$name2status;
 
     return 0 unless @names;     # No new names to check
 
     # Grok the list of all files in the repository at $commit
-    my @ls_files = split /\0/, $git->run(qw/ls-tree -r -z --name-only --full-tree/,
-                                         $commit eq ':0' ? 'HEAD' : $commit);
+    my @ls_files = split(
+        /\0/,
+        $git->run(qw/ls-tree -r -z --name-only --full-tree/,
+                  $commit ne ':0' ? $commit : $git->get_head_or_empty_tree),
+    );
 
     my $errors = 0;
 
@@ -291,7 +294,7 @@ EOS
 }
 
 sub deny_token {
-    my ($git, $ctx, $new_commit, $old_commit) = @_;
+    my ($git, $ctx, $commit) = @_;
 
     my $regex = $git->get_config($CFG => 'deny-token')
         or return 0;
@@ -304,23 +307,13 @@ EOS
         return 1;
     }
 
-    my @diff;
-
-    if (! defined $new_commit) {
-        die 'Internal error: $new_commit must be defined';
-    } elsif ($new_commit eq ':0') {
-        @diff = $git->run(qw/diff-index -p --diff-filter=AM --ignore-submodules/,
-                          "-G$regex", 'HEAD');
-    } elsif (! defined $old_commit) {
-        @diff = $git->run(qw/diff-tree -p --diff-filter=AM --ignore-submodules/,
-                          "-G$regex", $new_commit);
-    } else {
-        @diff = $git->run(qw/diff-tree -p --diff-filter=AM --ignore-submodules/,
-                          "-G$regex", $old_commit, $new_commit);
-    }
-
     # Extract only the lines showing addition of the $regex
-    @diff = grep {/^+.*?(?:$regex)/} @diff;
+    my @diff = grep {/^+.*?(?:$regex)/}
+        ($commit ne ':0'
+         ? $git->run(qw/diff-tree  -p --diff-filter=AM --ignore-submodules/,
+                     "-G$regex", $commit)
+         : $git->run(qw/diff-index -p --diff-filter=AM --ignore-submodules/,
+                     "-G$regex", $git->get_head_or_empty_tree));
 
     if (@diff) {
         $git->fault(<<EOS, {%$ctx, option => 'deny-token', details => join("\n", @diff)});
@@ -341,7 +334,7 @@ my %ACTION = (
 );
 
 sub check_acls {
-    my ($git, $ctx, $commit, $name_status) = @_;
+    my ($git, $ctx, $name2status) = @_;
 
     my @acls = eval { $git->grok_acls($CFG, 'AMD') };
     if ($@) {
@@ -349,16 +342,15 @@ sub check_acls {
         return 1;
     }
 
-    # Keep only ACLs matching the current user
-    @acls = grep {! exists $_->{who} || $git->match_user($_->{who})} @acls;
-
     return 0 unless @acls;
 
-    my $errors = 0;
+    # Collect the ACL errors and group them by ACL/ACTION so that we can produce
+    # more compact error messages.
+    my %acl_errors;
 
   FILE:
-    foreach my $file (sort keys %$name_status) {
-        my $statuses = $name_status->{$file};
+    foreach my $file (sort keys %$name2status) {
+        my $statuses = $name2status->{$file};
         foreach my $acl (@acls) {
             next unless ref $acl->{spec} ? $file =~ $acl->{spec} : $file eq $acl->{spec};
 
@@ -369,33 +361,64 @@ sub check_acls {
             next unless any {index($acl->{action}, $_) != -1} split //, $statuses;
 
             unless ($acl->{allow}) {
-                ++$errors;
-                my $myself = $git->authenticated_user();
                 my $action = $ACTION{$statuses} || $statuses;
-                $git->fault(<<EOS, {%$ctx, option => 'acl'});
-Authorization error: you ($myself) cannot $action this file:
-
-  $file
-
-Please, check your acl options.
-EOS
+                push @{$acl_errors{$acl->{acl}}{$action}}, $file;
             }
 
             next FILE;
         }
     }
 
-    return $errors;
+    if (%acl_errors) {
+        my $myself = $git->authenticated_user();
+        my %context = (%$ctx, option => 'acl');
+        while (my ($acl, $actions) = each %acl_errors) {
+            while (my ($action, $files) = each %$actions) {
+                my $these_files = scalar(@$files) > 1 ? 'these files' : 'this file';
+                $git->fault(<<EOS, \%context);
+Authorization error: you ($myself) cannot $action $these_files:
+
+  @{[join("\n  ", @$files)]}
+
+Due to the following acl:
+
+  $acl
+EOS
+            }
+        }
+    }
+
+    return scalar %acl_errors;
 }
 
 sub check_everything {
-    my ($git, $ctx, $old_commit, $new_commit, $name_status) = @_;
+    my ($git, $ref, $commit, $extra) = @_;
+
+    # The $extra information was generated by the --name-status --cc options to
+    # git-log. It has one line for each file affected in the commit. Merge
+    # commits only show files with conflicts. The format is "<S>+\t<FILE>".  <S>
+    # is one letter indicating how the file was affected, as documented in the
+    # --diff-filter option. <FILE> is the file path since the repository root,
+    # without a leading slash.
+
+    my %name2status;
+
+    if (defined $extra) {
+        foreach (split /\n/, $extra) {
+            if (/^(?<status>[ACDMRTUXB0-9]+)\t(?<file>.+)/) {
+                $name2status{$+{file}} = $+{status};
+            }
+        }
+    }
+
+    my %context = (ref => $ref);
+    $context{commit} = $commit unless $commit eq ':0';
 
     return
-        check_new_files(    $git, $ctx, $new_commit, $name_status) +
-        deny_case_conflicts($git, $ctx, $new_commit, $name_status) +
-        deny_token(         $git, $ctx, $new_commit, $old_commit) +
-        check_acls(         $git, $ctx, $new_commit, $name_status);
+        check_new_files($git, \%context, $commit, \%name2status) +
+        check_acls($git, \%context, \%name2status) +
+        deny_case_conflicts($git, \%context, $commit, \%name2status) +
+        deny_token($git, \%context, $commit);
 }
 
 # This routine can act both as an update or a pre-receive hook.
@@ -413,13 +436,15 @@ sub check_affected_refs {
 
         my ($old_commit, $new_commit) = $git->get_affected_ref_range($ref);
 
-        $errors += check_everything(
-            $git,
-            {ref => $ref},
+        my @commits = $git->get_commits(
             $old_commit,
             $new_commit,
-            $git->filter_name_status_in_range('ACMD', $old_commit, $new_commit),
+            [qw/--name-status --ignore-submodules -r --cc/],
         );
+
+        foreach my $commit (@commits) {
+            $errors += check_everything($git, $ref, $commit->commit, $commit->extra);
+        }
     }
 
     return $errors == 0;
@@ -436,11 +461,10 @@ sub check_commit {
 
     return 0 == check_everything(
         $git,
-        {},
-        ':0',
-        ':0',
-        $git->filter_name_status_in_index('ACMD'),
-    );
+        $current_branch,
+        ':0',                   # mark to signify the index
+        $git->run(qw/diff-index --name-status --ignore-submodules --no-commit-id --cached -r/,
+                  $git->get_head_or_empty_tree));
 }
 
 sub check_patchset {
@@ -459,15 +483,13 @@ sub check_patchset {
 
     return 1 unless $git->is_reference_enabled($branch);
 
-    my $commit = $opts->{'--commit'};
-
-    return 0 == check_everything(
-        $git,
-        {ref => $branch, commit => $commit},
-        "$commit^",
-        $commit,
-        $git->filter_name_status_in_commit('ACMD', $commit),
+    my ($commit) = $git->get_commits(
+        "$opts->{'--commit'}^",
+        $opts->{'--commit'},
+        [qw/-n 1 --name-status --ignore-submodules -r --cc/],
     );
+
+    return 0 == check_everything($git, $branch, $commit->commit, $commit->extra);
 }
 
 INIT: {
@@ -495,7 +517,7 @@ Git::Hooks::CheckFile - Git::Hooks plugin for checking files
 
 =head1 VERSION
 
-version 2.9.2
+version 2.9.3
 
 =head1 SYNOPSIS
 
